(*
Delphi 7 unit for the libfaad2 Wrapper Filter (LWF)
Copyright 2007-2008 Griga (http://www.dvbviewer.com)
Contact me via http://www.dvbviewer.info/forum/

This file is part of the LWF 1.1.3 source code.

    The LWF is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    The LWF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the LWF. If not, see <http://www.gnu.org/licenses/>.
*)

unit ULibFaad2;
//contains a wrapper class for the libfaad2.dll interface

interface

uses
  Windows;

const
  // object types for AAC
  MAIN      =  1;
  LC        =  2;
  SSR       =  3;
  LTP       =  4;
  HE_AAC    =  5;
  ER_LC     = 17;
  ER_LTP    = 19;
  LD        = 23;
  DRM_ER_LC = 27; // special object type for DRM

  // header types
  RAW       =  0;
  ADIF      =  1;
  ADTS      =  2;

  // SBR signalling
  NO_SBR           = 0;
  SBR_UPSAMPLED    = 1;
  SBR_DOWNSAMPLED  = 2;
  NO_SBR_UPSAMPLED = 3;

  // library output formats
  FAAD_FMT_16BIT   = 1;
  FAAD_FMT_24BIT   = 2;
  FAAD_FMT_32BIT   = 3;
  FAAD_FMT_FLOAT   = 4;
  FAAD_FMT_FIXED   = FAAD_FMT_FLOAT;
  FAAD_FMT_DOUBLE  = 5;

  // Capabilities
  LC_DEC_CAP            = $01; // Can decode LC
  MAIN_DEC_CAP          = $02; // Can decode MAIN
  LTP_DEC_CAP           = $04; // Can decode LTP
  LD_DEC_CAP            = $08; // Can decode LD
  ERROR_RESILIENCE_CAP  = $10; // Can decode ER
  FIXED_POINT_CAP       = $20; // Fixed point

  // Channel definitions
  FRONT_CHANNEL_CENTER = 1;
  FRONT_CHANNEL_LEFT   = 2;
  FRONT_CHANNEL_RIGHT  = 3;
  SIDE_CHANNEL_LEFT    = 4;
  SIDE_CHANNEL_RIGHT   = 5;
  BACK_CHANNEL_LEFT    = 6;
  BACK_CHANNEL_RIGHT   = 7;
  BACK_CHANNEL_CENTER  = 8;
  LFE_CHANNEL          = 9;
  UNKNOWN_CHANNEL      = 0;

  // DRM channel definitions
  DRMCH_MONO           = 1;
  DRMCH_STEREO         = 2;
  DRMCH_SBR_MONO       = 3;
  DRMCH_SBR_STEREO     = 4;
  DRMCH_SBR_PS_STEREO  = 5;

  // A decode call can eat up to FAAD_MIN_STREAMSIZE bytes per decoded channel,
  // so at least so much bytes per channel should be available in this stream
  FAAD_MIN_STREAMSIZE = 768; // 6144 bits/channel

type
  TNeAACDecHandle = Pointer;

  Tmp4AudioSpecificConfig = record
    //Audio Specific Info
    objectTypeIndex: Byte;
    samplingFrequencyIndex: Byte;
    samplingFrequency: DWord;
    channelsConfiguration: Byte;
    //GA Specific Info
    frameLengthFlag: Byte;
    dependsOnCoreCoder: Byte;
    coreCoderDelay: Word;
    extensionFlag: Byte;
    aacSectionDataResilienceFlag: Byte;
    aacScalefactorDataResilienceFlag: Byte;
    aacSpectralDataResilienceFlag: Byte;
    epConfig: Byte;

    sbr_present_flag: Byte;
    forceUpSampling: Byte;
    downSampledSBR: Byte;
  end;
  Pmp4AudioSpecificConfig = ^Tmp4AudioSpecificConfig;

  TNeAACDecConfiguration = record
    defObjectType: Byte;
    defSampleRate: DWord;
    outputFormat: Byte;
    downMatrix: Byte;
    useOldADTSFormat: Byte;
    dontUpSampleImplicitSBR: Byte;
  end;
  PNeAACDecConfiguration = ^TNeAACDecConfiguration;

  TNeAACDecFrameInfo = record
    bytesconsumed: DWord;
    samples: DWord;
    channels: Byte;
    error: Byte;
    samplerate: DWord;
    // SBR: 0: off, 1: on; upsample, 2: on; downsampled, 3: off; upsampled
    sbr: Byte;
    // MPEG-4 ObjectType
    object_type: Byte;
    //AAC header type; MP4 will be signalled as RAW also
    header_type: Byte;
    //multichannel configuration
    num_front_channels: Byte;
    num_side_channels: Byte;
    num_back_channels: Byte;
    num_lfe_channels: Byte;
    channel_position: array[0..63] of Byte;
    // PS: 0: off, 1: on
    ps: Byte;
  end;
  PNeAACDecFrameInfo = ^TNeAACDecFrameInfo;

  TNeAACDecGetErrorMessage = function(errcode: Byte): PChar; cdecl;
  TNeAACDecGetCapabilities = function: DWord; cdecl;
  TNeAACDecOpen = function: TNeAACDecHandle; cdecl;
  TNeAACDecGetCurrentConfiguration = function(hDecoder: TNeAACDecHandle):
    PNeAACDecConfiguration; cdecl;
  TNeAACDecSetConfiguration = function(hDecoder: TNeAACDecHandle;
    config: PNeAACDecConfiguration): Byte; cdecl;
  //Init the library based on info from the AAC file (ADTS/ADIF)
   TNeAACDecInit = function(hDecoder: TNeAACDecHandle;
    buffer: Pointer; buffer_size: Integer;
    var samplerate: DWord; var channels: Byte): Integer; cdecl;
  // Init the library using a DecoderSpecificInfo
  TNeAACDecInit2 = function(hDecoder: TNeAACDecHandle;
    pBuffer: Pointer; SizeOfDecoderSpecificInfo: Integer;
    var samplerate: DWord; var channels: Byte): Integer; cdecl;
  // Init the library for DRM
  TNeAACDecInitDRM = function(hDecoder: TNeAACDecHandle;
    samplerate: DWord; channels: Byte): Byte; cdecl;
  TNeAACDecPostSeekReset = procedure(hDecoder: TNeAACDecHandle; frame: DWord); cdecl;
  TNeAACDecClose = procedure(hDecoder: TNeAACDecHandle); cdecl;
  TNeAACDecDecode = function(hDecoder: TNeAACDecHandle;
    hInfo: PNeAACDecFrameInfo; buffer: Pointer; buffer_size: DWord): Pointer; cdecl;
  TNeAACDecDecode2 = function(hDecoder: TNeAACDecHandle;
    hInfo: PNeAACDecFrameInfo; buffer: Pointer; buffer_size: DWord;
    sample_buffer: Pointer; sample_buffer_size: DWord): Pointer; cdecl;
  TNeAACDecAudioSpecificConfig = function(pBuffer: Pointer; buffer_size: DWord;
    mp4ASC: Pmp4AudioSpecificConfig): Byte; cdecl;

  TLibFaad2 = class
  private
    hLib: THandle;
    hDec: TNeAACDecHandle;
    NeAACDecGetErrorMessage: TNeAACDecGetErrorMessage; //
    NeAACDecGetCapabilities: TNeAACDecGetCapabilities;
    NeAACDecOpen: TNeAACDecOpen; //
    NeAACDecGetCurrentConfiguration: TNeAACDecGetCurrentConfiguration; //
    NeAACDecSetConfiguration: TNeAACDecSetConfiguration; //
    NeAACDecInit: TNeAACDecInit; //
    NeAACDecInit2: TNeAACDecInit2;  //
    NeAACDecInitDRM: TNeAACDecInitDRM;
    NeAACDecPostSeekReset: TNeAACDecPostSeekReset; //
    NeAACDecClose: TNeAACDecClose; //
    NeAACDecDecode: TNeAACDecDecode; //
    NeAACDecDecode2: TNeAACDecDecode2;
    NeAACDecAudioSpecificConfig: TNeAACDecAudioSpecificConfig;  //
    function GetIsOpen: Boolean;
  public
    constructor Create(LibPath: String);
    destructor Destroy; override;
    function Open: Boolean;
    procedure Close;
    function GetErrorMessage(errcode: Byte): PChar;
    function GetCapabilities: DWord;
    function GetCurrentConfiguration: PNeAACDecConfiguration;
    function SetConfiguration(config: PNeAACDecConfiguration): Byte;
    //Init the library based on info from the AAC file (ADTS/ADIF)
    function Init(buffer: Pointer; buffer_size: DWord;
      var samplerate: DWord; var channels: Byte): Integer;
    // Init the library using a DecoderSpecificInfo
    function Init2(pBuffer: Pointer; SizeOfDecoderSpecificInfo: DWord;
      var samplerate: DWord; var channels: Byte): Integer;
    // Init the library for DRM
    function InitDRM(samplerate: DWord; channels: Byte): Byte;
    procedure PostSeekReset(frame: DWord);
    function Decode(out hInfo: TNeAACDecFrameInfo; buffer: Pointer;
      buffer_size: DWord): Pointer;
    function Decode2(out hInfo: TNeAACDecFrameInfo; buffer: Pointer;
      buffer_size: DWord; sample_buffer: Pointer; sample_buffer_size: DWord): Pointer;
    function AudioSpecificConfig(pBuffer: Pointer; buffer_size: DWord;
      mp4ASC: Pmp4AudioSpecificConfig): Byte;

    property IsOpen: Boolean read GetIsOpen;
  end;

implementation

uses
  Registry;

const
  sDLLName = 'libfaad2.dll';
  sFuncName: array[0..12] of String =
    ('NeAACDecGetErrorMessage','NeAACDecGetCapabilities','NeAACDecOpen',
     'NeAACDecGetCurrentConfiguration','NeAACDecSetConfiguration',
     'NeAACDecInit','NeAACDecInit2','NeAACDecInitDRM',
     'NeAACDecPostSeekReset','NeAACDecClose','NeAACDecDecode','NeAACDecDecode2',
     'NeAACDecAudioSpecificConfig');

constructor TLibFaad2.Create(LibPath: String);
var
  i: Integer;
  PP: PPointer;
begin
  inherited Create;
  LibPath := LibPath + sDLLName;
  hLib := LoadLibrary(PChar(LibPath));
  if hLib = 0 then
    hLib := LoadLibrary(sDLLName);
  if hLib <> 0 then
  begin
    PP := @@NeAACDecGetErrorMessage;
    for i := Low(sFuncName) to High(sFuncName) do
    begin
      PP^ := GetProcAddress(hLib,PChar(sFuncName[i]));
      Inc(PP);
    end;
  end;
end;

destructor TLibFaad2.Destroy;
begin
  Close;
  if hLib <> 0 then
    FreeLibrary(hLib);
  inherited;
end;

function TLibFaad2.GetIsOpen: Boolean;
begin
  result := assigned(hDec);
end;

function TLibFaad2.Open: Boolean;
begin
  if not assigned(hDec) and assigned(NeAACDecOpen) then
    hDec := NeAACDecOpen;
  result := assigned(hDec);
end;

procedure TLibFaad2.Close;
begin
  if assigned(hDec) and assigned(NeAACDecClose) then
  begin
    NeAACDecClose(hDec);
    hDec := nil;
  end;
end;

function TLibFaad2.GetErrorMessage(errcode: Byte): PChar;
begin
  result := nil;
  if assigned(NeAACDecGetErrorMessage) then
    result := NeAACDecGetErrorMessage(errcode);
end;

function TLibFaad2.GetCapabilities: DWord;
begin
  result := 0;
  if assigned(NeAACDecGetCapabilities) then
    result := NeAACDecGetCapabilities;
end;

function TLibFaad2.GetCurrentConfiguration: PNeAACDecConfiguration;
begin
  result := nil;
  if assigned(NeAACDecGetCurrentConfiguration) and assigned(hDec) then
    result := NeAACDecGetCurrentConfiguration(hDec);
end;

function TLibFaad2.SetConfiguration(config: PNeAACDecConfiguration): Byte;
begin
  result := 0;
  if assigned(NeAACDecSetConfiguration) and assigned(hDec) then
    result := NeAACDecSetConfiguration(hDec,config);
end;

function TLibFaad2.Init(buffer: Pointer; buffer_size: DWord;
  var samplerate: DWord; var channels: Byte): Integer;
begin
  result := 0;
  if assigned(NeAACDecInit) and assigned(hDec) then
    result := NeAACDecInit(hDec,buffer,buffer_size,samplerate,channels);
end;

function TLibFaad2.Init2(pBuffer: Pointer; SizeOfDecoderSpecificInfo: DWord;
  var samplerate: DWord; var channels: Byte): Integer;
begin
  result := 0;
  if assigned(NeAACDecInit2) and assigned(hDec) then
    result := NeAACDecInit2(hDec,pBuffer,SizeOfDecoderSpecificInfo,samplerate,channels);
end;

function TLibFaad2.InitDRM(samplerate: DWord; channels: Byte): Byte;
begin
  result := 0;
  if assigned(NeAACDecInitDRM) and assigned(hDec) then
    result := NeAACDecInitDRM(hDec,samplerate,channels);
end;

procedure TLibFaad2.PostSeekReset(frame: DWord);
begin
  if assigned(NeAACDecPostSeekReset) and assigned(hDec) then
    NeAACDecPostSeekReset(hDec,frame);
end;

function TLibFaad2.Decode(out hInfo: TNeAACDecFrameInfo; buffer: Pointer;
  buffer_size: DWord): Pointer;
begin
  result := nil;
  if assigned(NeAACDecDecode) and assigned(hDec) then
    result := NeAACDecDecode(hDec,@hInfo,buffer,buffer_size);
end;

function TLibFaad2.Decode2(out hInfo: TNeAACDecFrameInfo; buffer: Pointer;
  buffer_size: DWord; sample_buffer: Pointer; sample_buffer_size: DWord): Pointer;
begin
  result := nil;
  if assigned(NeAACDecDecode2) and assigned(hDec) then
    result := NeAACDecDecode2(hDec,@hInfo,buffer,buffer_size,sample_buffer,sample_buffer_size);
end;

function TLibFaad2.AudioSpecificConfig(pBuffer: Pointer; buffer_size: DWord;
  mp4ASC: Pmp4AudioSpecificConfig): Byte;
begin
  result := 0;
  if assigned(NeAACDecAudioSpecificConfig) then
    result := NeAACDecAudioSpecificConfig(pBuffer,buffer_size,mp4ASC);
end;

end.
