(*
Delphi 7 unit for the libfaad2 Wrapper Filter (LWF)
Copyright 2007-2008 Griga (http://www.dvbviewer.com)
Contact me via http://www.dvbviewer.info/forum/

This file is part of the LWF 1.1.3 source code.

    The LWF is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    The LWF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the LWF. If not, see <http://www.gnu.org/licenses/>.
*)

unit UWrapper;
//contains the Tfaad2Wrapper class

interface

uses
  Windows, SysUtils, MMSystem, ActiveX,
  DirectShow9, BaseClass, DSUtil,
  //DirectShow9, BaseClass and DSUtil are part of the DSPack (Borland Delphi API for Direct Show)
  //Available from http://www.progdigy.com
  UFilterData, ULibFaad2, UParser;

type
  TAACMode = (amADTS,amRaw,amLATM);

  Tfaad2Wrapper = class(TBCTransformFilter,IPersist,IPersistStream)
  private
    Lib: TLibFaad2;
    parser: TAACParser;
    data_size: Integer;
    data_buffer: PByteArray;
    OutSamplerate: DWord;
    OutChannels: Byte;
    ConfigData: TConfigData;
    aac_mode: TAACMode;
    is_reset: Boolean;
    last_Discontinuity: Boolean;
    specific_info_data: array[0..63] of Byte;
    specific_info_len: Integer;
    //small PTS FIFO queue
    nPTS: Integer;
    PTS_queue: array[0..2] of REFERENCE_TIME;

    procedure StoreConfig(ExtraData: Pointer; ExtraDataSize: Integer);
    function Decode(Data: PByte; Size: Integer): HRESULT;
    function Reset: Boolean;
    procedure PushPTS(const Value: REFERENCE_TIME);
    function PopPTS: REFERENCE_TIME;
    procedure CreateMediaType(var mt: TAMMediaType; var wfex: TWaveFormatExtensible);
  public
    constructor Create(const ObjName: string; Unk: IUnKnown);
    constructor CreateFromFactory(Factory: TBCClassFactory;
      const Controller: IUnknown); override;
    destructor Destroy; override;
    function DecideBufferSize(Allocator: IMemAllocator; prop: PAllocatorProperties): HRESULT; override;
    function CheckInputType(mtIn: PAMMediaType): HRESULT; override;
    function CheckTransform(mtIn, mtOut: PAMMediaType): HRESULT; override;
    function GetMediaType(Position: integer; out MediaType: PAMMediaType): HRESULT; override;
    function SetMediaType(direction: TPinDirection; pmt: PAMMediaType): HRESULT; override;
    function Receive(Sample: IMediaSample): HRESULT; override;
    function EndFlush: HRESULT; override;
    function Stop: HRESULT; override;
    //IPersistStream - for loading/saving of GRF files in GraphEdit
    function GetClassID(out classID: TCLSID): HResult; stdcall;
    function IsDirty: HResult; stdcall;
    function Load(const stm: IStream): HResult; stdcall;
    function Save(const stm: IStream; fClearDirty: BOOL): HResult; stdcall;
    function GetSizeMax(out cbSize: Largeint): HResult; stdcall;
  end;

implementation

uses
  Registry;

function Get_ObjectPath(CLSID: String): String;
//returns the file path of the filter with CLSID (if registered)
var
  reg: TRegistry;
begin
  reg := TRegistry.Create(KEY_READ);
  try
    try
      CLSID := 'CLSID\' + CLSID + '\InprocServer32';
      reg.RootKey := HKEY_CLASSES_ROOT;
      reg.OpenKeyReadOnly(CLSID);
      result := ExtractFilePath(Reg.ReadString(''));
    finally
      reg.free;
    end;
  except
    result := '';
  end;
end;

constructor Tfaad2Wrapper.Create(const ObjName: string; Unk: IUnKnown);
begin
  IsMultiThread := true;
  inherited Create(ObjName, Unk, CLSID_faad2Wrapper);
  GetMem(data_buffer,64*1024);
  //tell the TLibFaad2 class where this filter is stored, so it can load the dll from there
  Lib := TLibFaad2.Create(Get_ObjectPath(sCLSID_faad2Wrapper));
  parser := TAACParser.Create;
end;

constructor Tfaad2Wrapper.CreateFromFactory(Factory: TBCClassFactory;
  const Controller: IUnknown);
begin
  Create(Factory.Name, Controller);
end;

destructor Tfaad2Wrapper.Destroy;
begin
  Lib.Free;
  FreeAndNil(parser);
  if assigned(data_buffer) then
    FreeMem(data_buffer);
  inherited;
end;

function Tfaad2Wrapper.DecideBufferSize(Allocator: IMemAllocator; prop: PAllocatorProperties): HRESULT;
var
  Actual: ALLOCATOR_PROPERTIES;
begin
	Prop.cbBuffer	:= 64 * 1024;
	Prop.cBuffers	:= 3;
	result := Allocator.SetProperties(Prop^,Actual);
	if SUCCEEDED(result) and (Actual.cbBuffer < Prop.cbBuffer) then
    result := E_INVALIDARG;
end;

function Tfaad2Wrapper.CheckInputType(mtIn: PAMMediaType): HRESULT;
var
  i: Integer;
begin
	if IsEqualGuid(mtIn.majortype,MEDIATYPE_Audio) and
     IsEqualGuid(mtIn.formattype,FORMAT_WaveFormatEx) then
    for i := 0 to High(AACPinTypes) do
      if IsEqualGuid(mtIn.subtype,AACPinTypes[i].clsMinorType^) then
      begin
        result := S_OK;
        exit;
      end;
  result := VFW_E_TYPE_NOT_ACCEPTED;
end;

function Tfaad2Wrapper.CheckTransform(mtIn, mtOut: PAMMediaType): HRESULT;
begin
	result := CheckInputType(mtIn);
  if SUCCEEDED(result) and
     (not IsEqualGuid(mtOut.majortype,MEDIATYPE_Audio) or
      not IsEqualGuid(mtOut.subtype,MEDIASUBTYPE_PCM)) then
    result := VFW_E_TYPE_NOT_ACCEPTED;
end;

procedure Tfaad2Wrapper.CreateMediaType(var mt: TAMMediaType; var wfex: TWaveFormatExtensible);
const
  cChannelMask: array[0..6] of DWord =
    (SPEAKER_ALL,
     KSAUDIO_SPEAKER_MONO, KSAUDIO_SPEAKER_STEREO,
     KSAUDIO_SPEAKER_STEREO or SPEAKER_FRONT_CENTER, KSAUDIO_SPEAKER_QUAD,
     KSAUDIO_SPEAKER_QUAD or SPEAKER_FRONT_CENTER, KSAUDIO_SPEAKER_5POINT1);
begin
  FillChar(wfex,SizeOf(wfex),0);
  mt := OutputType; //template declared as const in UFilterData
  mt.pbFormat := @wfex;

  with wfex,Format do
  begin
    wBitsPerSample := 16;
    nChannels := OutChannels;
    nSamplesPerSec := OutSamplerate;
    nBlockAlign := OutChannels * (wBitsPerSample div 8);
    nAvgBytesPerSec := nBlockAlign * OutSamplerate;
    if OutChannels <= 2 then
    begin
      mt.cbFormat := SizeOf(TWaveFormatEx);
      cbSize := 0;
      wFormatTag := WAVE_FORMAT_PCM;
    end
    else begin //use WaveFormatExtensible for more than two channels
      mt.cbFormat := SizeOf(TWaveFormatExtensible);
      cbSize := SizeOf(TWaveFormatExtensible) - SizeOf(TWaveFormatEx);
      wFormatTag := WAVE_FORMAT_EXTENSIBLE;
      Samples.wValidBitsPerSample := wBitsPerSample;
      dwChannelMask := cChannelMask[Low(cChannelMask)]; //default = SPEAKER_ALL (?)
      if OutChannels in [Low(cChannelMask)..High(cChannelMask)] then
        dwChannelMask := cChannelMask[OutChannels];
      SubFormat := KSDATAFORMAT_SUBTYPE_PCM;
    end;
  end;
end;

function Tfaad2Wrapper.GetMediaType(Position: Integer; out MediaType: PAMMediaType): HRESULT;
var
  mt: TAMMediaType;
  wfex: TWaveFormatExtensible;
begin
  if not assigned(MediaType) then
  begin
    result := E_POINTER;
    exit;
  end;
	if Position < 0 then
  begin
    result := E_INVALIDARG;
    exit;
  end;
  if Position > High(PCMPinTypes) then
  begin
    result := VFW_S_NO_MORE_ITEMS;
    exit;
  end;
  if not Lib.IsOpen or not FInput.IsConnected then
  begin
    result := E_FAIL;
    exit;
  end;
  CreateMediaType(mt,wfex);
  CopyMediaType(MediaType,@mt);
  result := NOERROR;
end;

function Tfaad2Wrapper.SetMediaType(direction: TPinDirection; pmt: PAMMediaType): HRESULT;
//The connection behaviour is as follows: If no extra data is appended to
//WaveFormatEx, the filter assumes ADTS AAC, regardless of the media subtype.
//if WaveFormatEx contains extra data, it depends on the media subtype:
//MEDIASUBTYPE_AAC: ADTS assumed, if a whole ADTS frame is attached (Elecard Demux),
//otherwise Raw AAC, if the extra data is accepted by faad2,
//MEDIASUBTYPE_LATM_AAC: LOAS/LATM AAC assumed, f the extra data is accepted by
//  the LATM parser, otherwise failure.
var
  ExtraData: Pointer;
  ExtraDataSize: DWord;
begin
  if not assigned(pmt) then
  begin
    result := E_POINTER;
    exit;
  end;
  result := inherited SetMediaType(direction,pmt);
  if FAILED(result) then
    exit;
  case direction of
    PINDIR_INPUT:
      begin
        lib.Close;
        specific_info_len := 0;
        aac_mode := amADTS; //default
        if IsEqualGuid(pmt.formattype,FORMAT_WaveFormatEx) and
           assigned(pmt.pbFormat) then
        begin
          //read settings from WaveFormatEx - we may get no extra data
          with PWaveFormatEx(pmt.pbFormat)^, ConfigData do
          begin
            Channels := nChannels;
            Samplerate := nSamplesPerSec;
            AudioObjectType := LC; //2 = low complexity
            specific_info_len := WriteConfig(@specific_info_data,ConfigData);
          end;
          ExtraDataSize := PWaveFormatEx(pmt.pbFormat).cbSize;
          if ExtraDataSize > 1 then //there's something meaningful attached to WaveFormatEx
          begin
            ExtraData := Pointer(DWord(pmt.pbFormat) + SizeOf(TWaveFormatEx));
            if IsEqualGUID(pmt.subtype,MEDIASUBTYPE_LATM_AAC) then
            //LATM config data are expected here, which means, a LOAS packet
            //encapsulating a LATM packet with use_same_mux flag = 0
            begin
              if not parser.ReadLATMData(ExtraData,ExtraDataSize) or
                (Length(parser.config) = 0) then //refuse connection
              begin
                result := E_FAIL;
                exit;
              end;
              //New extra data: audio_specific_config extracted by LATM parser
              with parser.config[0] do
              begin
                ExtraData := @extra;
                ExtraDataSize := extrasize;
              end;
              aac_mode := amLATM;
            end
            else if parser.ReadADTSData(ExtraData,ExtraDataSize) then
              //a whole ADTS frame is attached (Elecard Demux)
              with parser.config[0] do
              begin
                ExtraData := @extra;
                ExtraDataSize := extrasize;
              end
            else
              aac_mode := amRaw; //raw samples (without headers), no parsing required
            StoreConfig(ExtraData,ExtraDataSize);
          end; //if ExtraSize > 1
        end; //if IsEqualGuid(pmt.formattype,FORMAT_WaveFormatEx)
        is_reset := false;
        if not Reset then
          result := E_FAIL;
      end; //PINDIR_INPUT
    PINDIR_OUTPUT:
      begin
        if not IsEqualGuid(pmt.formattype,FORMAT_WaveFormatEx) then
        begin
          result := E_FAIL;
          exit;
        end;
        if not assigned(pmt.pbFormat) then
        begin
          result := E_POINTER;
          exit;
        end;
        with PWaveFormatEx(pmt.pbFormat)^ do
          if (nSamplesPerSec <> OutSamplerate) or (nChannels <> OutChannels) then
          begin
            result := E_FAIL;
            exit;
          end;
      end;
  end;
end;

procedure Tfaad2Wrapper.PushPTS(const Value: REFERENCE_TIME);
begin
  if nPTS >= Length(PTS_queue) then //prevent overflow
    PopPTS;
  PTS_Queue[nPTS] := Value;
  Inc(nPTS);
end;

function Tfaad2Wrapper.PopPTS: REFERENCE_TIME;
begin
  if nPTS <= 0 then
    result := InvalidClock
  else begin
    result := PTS_queue[0];
    Dec(nPTS);
    if nPTS > 0 then
      move(PTS_queue[nPTS],PTS_queue[nPTS-1],nPTS*SizeOf(REFERENCE_TIME));
  end;
end;

procedure Tfaad2Wrapper.StoreConfig(ExtraData: Pointer; ExtraDataSize: Integer);
begin
  ReadConfig(ExtraData,ConfigData); //extract AudioObjectType, Samplerate, Channels
  if ExtraDataSize > SizeOf(specific_info_data) then
    ExtraDataSize := SizeOf(specific_info_data);
  specific_info_len := ExtraDataSize;
  Move(ExtraData^,specific_info_data,ExtraDataSize);
end;

function Tfaad2Wrapper.Reset: Boolean;
begin
  result := true;
  if not is_reset then // flush data, reset faad2
  begin
    data_size := 0;
    nPTS := 0;
    last_Discontinuity := true;
    Lib.Close;
    result := (specific_info_len > 0) and
              Lib.Open and
              (Lib.Init2(@specific_info_data,specific_info_len,OutSamplerate,OutChannels) >= 0);
    is_reset := true;
  end;
end;

function Tfaad2Wrapper.Stop: HRESULT;
begin
  result := inherited Stop;
  Reset;
end;

function Tfaad2Wrapper.EndFlush: HRESULT;
begin
  Reset;
  result := inherited EndFlush;
end;

function Tfaad2Wrapper.Receive(Sample: IMediaSample): HRESULT;
var
  Data: PByte;
  pmt: PAMMediaType;
  Size, DstSize, ret, data_pos: Integer;
  rtStart,rtStop: REFERENCE_TIME;
  tmpBuf: array[0..4*1024-1] of Byte;
  s: DWord;
  c: Byte;
begin
  if Sample.IsDiscontinuity = S_OK then
    //flush buffers and reset faad2, so we don't get nasty digital noise
    Reset;
  if not Lib.IsOpen then
  begin
    Sleep(1);
    result := E_FAIL;
    exit;
  end;

  //get the sample and the PTS
	result := Sample.GetPointer(Data);
  if FAILED(result) then
    exit;
	Size := Sample.GetActualDataLength;
	if SUCCEEDED(Sample.GetTime(rtStart,rtStop)) then
    PushPTS(rtStart);
  result := NOERROR;

  if aac_mode = amRaw then //no headers to parse
  begin
    //GetMediaType is the only way to detect a config change in raw mode
    //The demuxer must take care of it
    if Sample.GetMediaType(pmt) = S_OK then
    begin
      SetMediaType(PINDIR_INPUT,pmt);
      DeleteMediaType(pmt);
    end;
    result := Decode(Data,Size);
    if FAILED(result) then
      Reset;
    exit;
  end;

  //Put the sample to the parser buffer. The following code does not rely on a
  //particular audio data alignment or size (e.g. one complete packet per sample).
  //It collects everything in the buffer and does the necessary parsing. Thus the
  //stream may start at any point and arrive in random portions.
  data_pos := 0;
  Move(Data^,data_buffer[data_size],Size);
  Inc(data_size,Size);
  DstSize := 0;

  repeat
    if aac_mode = amLATM then //search LOAS/LATM packet, extract raw AAC payload
      ret := parser.ReadLATMPacket(@data_buffer[data_pos],data_size,@tmpBuf,DstSize)
    else //search and extract ADTS packet (with header, faad2 may need to read it)
      ret := parser.ReadADTSPacket(@data_buffer[data_pos],data_size,@tmpBuf,DstSize);
    //ret = number of consumed bytes
    if ret > 0 then
    begin
      if DstSize > 0 then
      begin
        with parser.config[0] do
          if not ConfigEqual(@Extra,@specific_info_data) then
          begin //change of audio_specific_configuration
            StoreConfig(@Extra,ExtraSize);
            //resetting libfaad2 seems to be the only way to handle a 5.1 -> 2.0 change
            //though 2.0 -> 5.1 works without it. Strange...
            Lib.Close;
            if Lib.Open then
              Lib.Init2(@Extra,ExtraSize,s,c);
            is_reset := true;
          end;
        result := Decode(@tmpbuf,DstSize); //decode the data and send it downstream
        if FAILED(result) then
        begin
          Sleep(1);
          EndOfStream;
          //Reset;
          exit;
        end;
      end;
      Inc(data_pos,ret);
      Dec(data_size,ret);
    end;
  until ret <= 0;
  //move rest to buffer start
  if (data_pos > 0) and (data_size > 0) then
    Move(data_buffer[data_pos],data_buffer[0],data_size);
end;

function Tfaad2Wrapper.Decode(Data: PByte; Size: Integer): HRESULT;
//faad2 and DirectShow use a different sample order if the number of channels is > 2
// faad order:  C, L, R, BL,  BR, LFE
// DShow order: L, R, C, LFE, BL, BR
//Thus the samples have to be mapped to the correct channel by using the following table:
const
  cMaxChannels = 6;
  ChMap: array[0..cMaxChannels-1,0..cMaxChannels-1] of Integer =
  (
  (0, 0, 0, 0, 0, 0), //mono, no remapping
  (0, 1, 0, 0, 0, 0), // l, r -> l, r, no remapping
  (2, 0, 1, 0, 0, 0), // c ,l, r -> l, r, c
  (2, 0, 1, 3, 0, 0), // c, l, r, bc -> l, r, c, bc
  (2, 0, 1, 3, 4, 0), // c, l, r, bl, br -> l, r, c, bl, br
  (2, 0, 1, 4, 5, 3)  // c, l, r, bl, br, lfe -> l, r, c, lfe, bl, br
  );
var
  buf: PWord;
  info: TNeAACDecFrameInfo;
  sample: IMediaSample;
  start, stop: REFERENCE_TIME;
  ret, i, j, cd: Integer;
  mt: TAMMediaType;
  wfex: TWaveFormatExtensible;
begin
  result := NOERROR;
  //we fetch the PTS from the queue here, because the function may exit early
  //due to an erroneous packet, and we dont want PTS to get retained in this case
  start := PopPTS;
  if is_reset and (aac_mode = amADTS) then //let faad2 read the ADTS header
  begin
    ret := Lib.Init(Data,Size,info.samplerate,info.channels);
    if ret < 0 then
      exit;
    Inc(Data,ret);
    Dec(Size,ret);
  end;
  FillChar(Info,SizeOf(Info),0);
  buf := Lib.Decode(info,Data,Size); //decode the stuff and return some info about it
  if not assigned(buf) or (info.error <> 0) then
    //we don't return E_FAIL here, since we don't want a single bad packet to stop the whole show
    exit;

  is_reset := false;
  Size := info.samples * SizeOf(Word);

  //send the decoded data downstream
  result := Output.GetDeliveryBuffer(Sample,nil,nil,0);
  if FAILED(result) then
    exit;
  if (info.samplerate <> OutSamplerate) or (info.channels <> OutChannels) then //format change
  begin
    OutSamplerate := info.samplerate;
    OutChannels := info.channels;
    CreateMediaType(mt,wfex);
    Output.SetMediaType(@mt); //set the output pin media type
    sample.SetMediaType(@mt);
  end;
  sample.GetPointer(Data);

  if info.channels <= 2 then
    Move(buf^,data^,Size)
  else begin //channel remapping
    cd := info.channels - 1;
    for i := 0 to info.samples div info.channels - 1 do
    begin
      for j := 0 to cd do
      begin
        PWordArray(Data)[ChMap[cd,j]] := buf^;
        Inc(buf);
      end;
      Inc(Data,info.channels * SizeOf(Word));
    end;
  end;

  sample.SetActualDataLength(Size);
  if start <> InvalidClock then
  begin
    stop := start + 1;
    sample.SetTime(@start,@stop);
  end;
  sample.SetDiscontinuity(last_Discontinuity); //pass the discontinuity on
  last_Discontinuity := false;
  result := Output.Deliver(sample);
end;

// IPersistStream

function Tfaad2Wrapper.GetClassID(out classID: TCLSID): HResult; stdcall;
begin
  classID := CLSID_faad2Wrapper;
  result := S_OK;
end;

function Tfaad2Wrapper.IsDirty: HResult; stdcall;
begin
  result := S_FALSE;
end;

function Tfaad2Wrapper.Load(const stm: IStream): HResult; stdcall;
begin
  result := S_OK;
end;

function Tfaad2Wrapper.Save(const stm: IStream; fClearDirty: BOOL): HResult; stdcall;
begin
  result := S_OK;
end;

function Tfaad2Wrapper.GetSizeMax(out cbSize: Largeint): HResult; stdcall;
begin
  cbSize := 0;
  result := S_OK;
end;

initialization
  TBCClassFactory.CreateFilter(Tfaad2Wrapper, faad2WrapperName,
    CLSID_faad2Wrapper, CLSID_LegacyAmFilterCategory,
    MERIT_NORMAL, Length(AACPins), @AACPins);
end.
